#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov  3 18:48:59 2019
@author: Olivier THUAL

Will the water overflow ?
Data assimilation with time evolution
"""
#  clear all
for iglob in list(globals().keys()):
    if(iglob[0] != '_'):
        exec('del {}'.format(iglob))
# import libraries
import numpy as np
import matplotlib.pyplot as plt
        


def cG_nol(x):
    # Observation operator
    al=x[0,0]; P=x[1,0]; H0=x[2,0]
    y=(H0-P/al)*np.exp(-al*tobs)+P/al
    y=y.reshape(nobs,1)
    return y

def G_lin(x):
    # Observation operator
    al=x[0,0]; P=x[1,0]; H0=x[2,0]
    exp=np.exp(-al*tobs); Psal=P/al
    dGdal=(-(H0-Psal)*tobs+Psal**2)*exp-Psal**2
    dGdP=(1-exp)/al
    dGdH0=exp
    G=np.array([dGdal.T, dGdP.T, dGdH0.T]).T
    return G

def inifig(xaxe=0,yaxe=0,xlab='x',ylab='y'):
    plt.figure()
    plt.axvline(xaxe)
    plt.axhline(yaxe)
    plt.xticks(fontsize=16)
    plt.yticks(fontsize=16)
    plt.xlabel(xlab,fontsize=16 )
    plt.ylabel(ylab,fontsize=16)
    
def Jb_cost(x):
    Jb=(x-xb).T.dot(Binv).dot(x-xb)
    return Jb  
 
def Jr_cost(x):
    Jr=sum(yo-cG_nol(x))**2/(2*sigr**2)
    return Jr

def J_cost(x):
    return Jb_cost(x)+Jr_cost(x)

def Traj(x):
    al=x[0,0]; P=x[1,0]; H0=x[2,0]
    H=(H0-P/al)*np.exp(-al*t)+P/al
    return H
    
   
# Parameters
tmax=5; nt=100;
t=np.linspace(0,tmax,nt)
Hmin=-.2; Hmax=1.2

# True state   
alt=1; Pt=1; H0t=0;

xt=np.array([alt, Pt,H0t]).reshape(3,1)
Ht=Traj(xt)

# For background
sigal=.2; sigP=.2; sigH0=.1; rhoaP=0

# For measurements
sigr=.1;
nobs=3; dt=1; tobs=np.linspace(dt,nobs*dt,nobs)
yt=cG_nol(xt)
#print('True measurements\n',yt)

# Figure
inifig(xaxe=0,yaxe=0,xlab='t',ylab='H')
plt.title('True Trajectory',fontsize=16)
plt.plot(t,Ht,linewidth=3,color='blue')
plt.scatter(tobs,yt,marker='o',color='blue',s=100)
Hlt=Pt/alt; plt.plot([0,tmax],[Hlt,Hlt],'--b',linewidth=3)
plt.ylim(Hmin,Hmax) 
#plt.show()


# B and R matrice
B11=sigal**2; B22=sigP**2; B33=sigH0**2
B12=rhoaP*sigal*sigP
B=np.array([[B11, B12,   0],\
   [B12, B22,   0 ],\
   [  0,   0, B33]])
Binv = np.linalg.inv(B) 
R=sigr**2*np.eye(nobs,nobs)
Rinv = np.linalg.inv(R) 


# Scores for the estimation of sigma_a
if True :
    I=9;
    
    for i in range(I) :
        print('======= Experiment',i)
        
        # Background
        eps_al=np.random.normal(0,sigal)
        eps_P=np.random.normal(0,sigP)
        eps_H0=np.random.normal(0,sigH0)
        alb=alt+eps_al
        Pb=Pt+eps_P+eps_al*rhoaP*sigP/sigal
        H0b=H0t+eps_H0
        xb=np.array([[alb, Pb, H0b]]).reshape(3,1)
        Hb=Traj(xb)
        #print("Background\n",xb)
        
        # Measurments
        yo=yt+np.random.normal(0,sigr,nobs).reshape(nobs,1)
        d=yo-cG_nol(xb)
        #print("Innovation\n",d)
        
        # Analysis
        G=G_lin(xb);
        K=B@G.T@np.linalg.inv(G@B@G.T+R)
        xa=xb+K@d
        ala=xa[0,0]; Pa=xa[1,0]; H0a=xa[2,0]
        #print("Analysis\n",xa)
        
        # Plot Figure
        inifig(xaxe=0,yaxe=0,xlab='t',ylab='H')
        plt.title('Experiment '+str(i),fontsize=16)  
        
        # True trajectory
        plt.plot(t,Ht,linewidth=3,color='blue')
        plt.scatter(tobs,yo,marker='o',color='blue',s=100)
        Hlt=Pt/alt; plt.plot([0,tmax],[Hlt,Hlt],'--b',linewidth=3)
       
        # Background
        plt.plot(t,Hb,linewidth=3,color='black')
        Hlb=Pb/alb; plt.plot([0,tmax],[Hlb,Hlb],'--k',linewidth=3)
        
        # Analysis
        H=Traj(xa); plt.plot(t,H,linewidth=3,color='red')
        Hla=Pa/ala; plt.plot([0,tmax],[Hla,Hla],'--r',linewidth=3)
        
        # Bound and save
        plt.ylim(Hmin,Hmax) 
        plt.savefig('Traj'+str(i)+'.pdf')
        plt.show(); plt.close()
    
# Show all figures in command line model
print("Close all figures to stop the program if locked")

    
      




